
import java.net.*;
import java.io.*;
import java.nio.*;
import java.util.regex.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.text.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.border.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.nio.channels.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.net.*;
import javax.net.ssl.*;
import java.security.*;
import java.util.jar.*;
import java.security.cert.*;
import java.math.*;

import org.jdom.*;
import org.jdom.adapters.*;
import org.jdom.filter.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.transform.*;
import org.jdom.xpath.*;

public class RuleRunner {
    public class Rule {
        public String name;
            
        public int patternStart;
        public int patternEnd;
        public String pattern;
        public int patternWidth;
        public int patternHeight;
        public String regex;
        public Pattern compiledRegex;
            
        public int uses = 0;
            
        public class MatchGroup {
            public char symbol;
            public String name;
            public String plural;
            public String regex;
                
            public Set<Spot> spots = new HashSet<Spot>();
            public boolean clickThese = false;
            public boolean markThese = false;
                
            public int unexploredCount = 0;
            public int mineCount = 0;
            
            public MatchGroup(char symbol, String name, String plural, String regex) {
                this.symbol = symbol;
                this.name = name;
                this.plural = plural;
                this.regex = regex;
            }
            
            public String extractClickAndMark(String rhs) {
                String a = rhs.replaceAll("click." + plural, "");
                if (!a.equals(rhs)) {
                    clickThese = true;
                }
                String b = a.replaceAll("mark." + plural, "");
                if (!b.equals(a)) {
                    markThese = true;
                }
                return b;
            }
            
            public void addMarksAndClicks(char c, Spot spot) {
                if (c == symbol) {
                    spots.add(spot);
                    
                    if (clickThese) {
                        Rule.this.clickThese.add(spot);
                    }
                    if (markThese) {
                        Rule.this.markThese.add(spot);
                    }
                }
            }
            
            public void match(int x, int y, String stringBoard) {
                unexploredCount = 0;
                mineCount = 0;
                
                for (Spot s : spots) {
                    switch (stringBoard.charAt(
                        (x + s.x) + (y + s.y) * stringBoardWidth))
                    {
                        case '?' : unexploredCount++; break;
                        case 'M' : mineCount++; break;
                    }
                }
                
                js.setGlobalValue(name + "-unknown-count", unexploredCount);
                js.setGlobalValue(name + "-mine-count", mineCount);
            }
            
            public void addVariables(Vector<String> variableList) {
                variableList.add(name + "-unknown-count");
                variableList.add(name + "-mine-count");
            }
            
            public void appendMarkClickCommands(StringBuffer buf) {
                if (markThese) {
                    buf.append("mark-" + plural + "\n");
                }
                if (clickThese) {
                    buf.append("click-" + plural + "\n");
                }
            }
        }
        public Vector<MatchGroup> matchGroups = new Vector<MatchGroup>();
        public Map<String, Spot> variables = new HashMap<String, Spot>();
            
        public int jsStart;
        public int jsEnd;
        public String jsString = null;
        public Object jsObject = null;
            
        public String markPattern;
        public Set<Spot> markThese = new HashSet<Spot>();
        public Set<Spot> clickThese = new HashSet<Spot>();
            
        public Rule(String s, int start, int end) {
            matchGroups.add(new MatchGroup('*', "star", "stars", "[0-9\\|\\?M]"));
            matchGroups.add(new MatchGroup('#', "pound", "pounds", "[0-9\\|\\?M]"));
            matchGroups.add(new MatchGroup('$', "dollar", "dollars", "[0-9\\|\\?M]"));
            matchGroups.add(new MatchGroup('+', "plus", "plusses", "[0-9\\|M]"));
            
            Matcher mm = Pattern.compile("(?msi)(.*?)^\\s*=>(.*)").matcher(s);
            mm.region(start, end);
            if (!mm.find()) {
                throw new IllegalArgumentException("missing '=>' on it's own line, see (" + start + ", " + end + "): " + s.substring(start, end));
            }
            
            Matcher mmm = Pattern.compile("(?msi)(.*?)\n\\s*(.*?)\\s*(\\(.*)?\\z").matcher(s);
            mmm.region(mm.start(1), mm.end(1));
            if (!mmm.find()) {
                throw new IllegalArgumentException("malformed antecedent, see (" + mm.start(1) + ", " + mm.end(1) + "): " + mm.group(1));
            }
            
            name = mmm.group(1);
            pattern = mmm.group(2);
            
            patternStart = mmm.start(2);
            patternEnd = mmm.end(2);
            
            String[] patternLines = pattern.split("\\s+");
            if (patternLines.length == 0) {
                throw new IllegalArgumentException("malformed pattern, see (" + patternStart + ", " + patternEnd + "): " + pattern);
            }
            
            patternWidth = patternLines[0].length();
            patternHeight = patternLines.length;
            
            if (patternWidth > biggestPatternSize) {
                biggestPatternSize = patternWidth;
            }
            if (patternHeight > biggestPatternSize) {
                biggestPatternSize = patternHeight;
            }
            
            for (String line : patternLines) {
                if (line.length() != patternWidth) {
                    throw new IllegalArgumentException("not all lines in this pattern are the same length, see (" + patternStart + ", " + patternEnd + "): " + pattern);
                }
            }
            
            if (mmm.group(3) != null) {
                jsStart = mmm.start(3);
                jsEnd = mmm.end(3);
                jsString = mmm.group(3);
                jsObject = null;
                try {
                    jsObject = js.read(jsString);
                } catch (Exception e) {
                    throwSchemeException(e);
                }
            }
            
            String rhs = mm.group(2);
            for (MatchGroup g : matchGroups) {
                rhs = g.extractClickAndMark(rhs);
                
                String lhsP = pattern.replaceAll("\\s+", "");
                int i = 0;
                for (int y = 0; y < patternHeight; y++) {
                    for (int x = 0; x < patternWidth; x++) {
                        char lhsC = lhsP.charAt(i++);
                        g.addMarksAndClicks(lhsC, new Spot(x, y));
                    }
                }
            }
            
            if (rhs.matches("(?s).*?\\S.*")) {
                if (rhs.matches("(?msi)\\s*(?:\\s*\\S{" + patternWidth + "}){" + patternHeight + "}\\s*")) {
                    markPattern = rhs;
                    String lhsP = pattern.replaceAll("\\s+", "");
                    String rhsP = markPattern.replaceAll("\\s+", "");
                    int i = 0;
                    for (int y = 0; y < patternHeight; y++) {
                        for (int x = 0; x < patternWidth; x++) {
                            char lhsC = lhsP.charAt(i);
                            char rhsC = rhsP.charAt(i);
                            
                            if (rhsC == 'M' && lhsC != 'M') {
                                markThese.add(new Spot(x, y));
                            } else if (rhsC == '-' || (rhsC >= '0' && rhsC <= '8')) {
                                if (!(lhsC == '-' || (lhsC >= '0' && lhsC <= '8'))) {
                                    clickThese.add(new Spot(x, y));
                                }
                            }
                            
                            i++;
                        }
                    }
                    if (markThese.size() == 0 && clickThese.size() == 0) {
                        throw new IllegalArgumentException("consequent pattern doesn't seem to mark any mines, or click any squares, see (" + mm.start(2) + ", " + mm.end(2) + "): " + mm.group(2));
                    }
                } else {
                    throw new IllegalArgumentException("consequent pattern size does not match antecedent, see (" + mm.start(2) + ", " + mm.end(2) + "): " + mm.group(2));
                }
            }
            if (markThese.size() == 0 && clickThese.size() == 0) {
                throw new IllegalArgumentException("Right hand side doesn't seem to mark any mines, or click any squares, see (" + mm.start(2) + ", " + mm.end(2) + "): " + mm.group(2));
            }
        }
            
        public void updateRegex() {
            StringBuffer buf = new StringBuffer();
            buf.append("(?s)");
            String p = pattern.replaceAll("\\s+", "");
            int i = 0;
            for (int y = 0; y < patternHeight; y++) {
                if (y != 0) {
                    buf.append(".{" + (stringBoardWidth - patternWidth) + "}");
                }
                for (int x = 0; x < patternWidth; x++) {
                    char ch = p.charAt(i++);
                    
                    if (ch == 'M')          buf.append(ch);
                    else if (ch == '|')     buf.append("\\|");
                    else if (ch >= 'a' && ch <= 'z') {
                        buf.append("\\d");
                        variables.put("" + ch, new Spot(x, y));
                    } else if (ch >= '0' && ch <= '8')
                                            buf.append(ch);
                    else if (ch == '-')     buf.append("[0-9\\|]");
                    else if (ch == '?')     buf.append("\\?");
                    else {
                        boolean found = false;
                        for (MatchGroup g : matchGroups) {
                            if (ch == g.symbol) {
                                buf.append(g.regex);
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            throw new IllegalArgumentException("unrecognized symbol in pattern: " + ch + ", see (" + patternStart + ", " + patternEnd + "): " + pattern);
                        }
                    }
                }
            }
            
            regex = buf.toString();
            compiledRegex = Pattern.compile(regex);
        }
        
        public boolean match(String stringBoard, int symmetricSpace) {
            Matcher m = compiledRegex.matcher(stringBoard);
            while (m.find()) {
                int x = m.start(0) % stringBoardWidth;
                int y = m.start(0) / stringBoardWidth;
                
                for (MatchGroup g : matchGroups) {
                    g.match(x, y, stringBoard);
                }
                
                for (String k : variables.keySet()) {
                    Spot spot = variables.get(k);
                    int value = Integer.parseInt("" + stringBoard.charAt(
                        (x + spot.x) + (y + spot.y) * stringBoardWidth));
                    
                    int mineCount = 0;
                    for (int oy = -1; oy <= 1; oy++) {
                        for (int ox = -1; ox <= 1; ox++) {
                            if (ox == 0 && oy == 0) continue;
                            if ('M' == stringBoard.charAt(
                                (x + spot.x + ox) + (y + spot.y + oy) * stringBoardWidth))
                            {
                                mineCount++;
                            }
                        }
                    }
                    
                    js.setGlobalValue(k, js.toObject(value));
                    js.setGlobalValue(k + "-mine-count", js.toObject(mineCount));
                }
                
                boolean found = false;
                try {
                    found = jsObject == null || js.eval(jsObject) == Boolean.TRUE;
                } catch (Exception e) {
                    throwSchemeException(e);
                }
                if (found) {
                    int X = x - (biggestPatternSize - 1);
                    int Y = y - (biggestPatternSize - 1);
                    
                    if (session.guess(markThese, clickThese, X, Y, symmetricSpace))
                    {
                        StringBuffer buf = new StringBuffer();
                        Spot s = board.toSymmetricSpace(new Spot(X, Y), symmetricSpace);
                        buf.append("rx = " + s.x + ", ry = " + s.y + ", rw = " + patternWidth + ", rh = " + patternHeight + ", s = " + symmetricSpace + "\n");
                        buf.append("\n");
                        buf.append("green rectangle matched:\n");
                        for (int oy = 0; oy < patternHeight; oy++) {
                            for (int ox = 0; ox < patternWidth; ox++) {
                                buf.append(stringBoard.charAt(
                                    (x + ox) + (y + oy) * stringBoardWidth));
                            }
                            buf.append("\n");
                        }
                        buf.append("\n");
                        buf.append("matched by rule: " + name + "\n");
                        buf.append(pattern.replaceAll("\\s+", "\n"));
                        buf.append("\n");
                        
                        if (jsString != null) {
                            Vector<String> variableList = new Vector<String>();
                            for (String k : variables.keySet()) {
                                variableList.add(k);
                            }
                            for (MatchGroup g : matchGroups) {
                                g.addVariables(variableList);
                            }
                            
                            String jsStringUnderscores = jsString.replaceAll("\\-", "_");
                            for (String v : variableList) {
                                if (jsStringUnderscores.matches("(?si).*?\\b" + v.replaceAll("\\-", "_") + "\\b.*")) {
                                    buf.append(v + " = " + js.getGlobalValue(v) + "\n");
                                }
                            }
                            buf.append("\n");
                            buf.append(jsString);
                        }
                        buf.append("\n=>\n");
                        if (markPattern != null) {
                            buf.append(markPattern.replaceAll("\\s+", "\n"));
                        }
                        for (MatchGroup g : matchGroups) {
                            g.appendMarkClickCommands(buf);
                        }
                        session.annotate(buf.toString());
                        
                        uses++;
                        
                        return true;
                    }
                }
                
                m.region(m.start(0) + 1, stringBoard.length());
            }
            return false;
        }
        
        public void throwSchemeException(Exception e) {
            String niceError = "" + e;
            Matcher mm = Pattern.compile("(?si).*?ERROR: (.*)").matcher(niceError);
            if (mm.find()) {
                niceError = mm.group(1);
            }
            throw new IllegalArgumentException("scheme error: " + niceError + ", see (" + jsStart + ", " + jsEnd + "): " + jsString);
        }
    }
    
    jscheme.JScheme js = null;
    public Vector<Rule> rules = new Vector<Rule>();
    int biggestPatternSize = 2; // this way we ensure at least a buffer of 1 tile around the board
    int stringBoardWidth = 0;
    
    Board board = null;
    GameSession session = null;
    
    public RuleRunner(String filename, Board board, GameSession session, ActionListener callback) throws Exception {
        this.board = board;
        this.session = session;
        
        js = new jscheme.JScheme();
        
        String s = u.slurp(filename);
        Matcher m = Pattern.compile("(?msi)^\\s*rule:(.*?)(^\\s*rule:|\\z)").matcher(s);
        while (m.find()) {
            rules.add(new Rule(s, m.start(1), m.end(1)));
            m.region(m.start(2), s.length());
        }
        stringBoardWidth = board.width + ((biggestPatternSize - 1) * 2) + 1;
        for (Rule r : rules) {
            r.updateRegex();
        }
        
        boolean progress = true;
        mainLoop : while (progress && !session.gameOver) {
            progress = false;
            
            for (int symmetricSpace = 0; symmetricSpace < 8; symmetricSpace++) {
                String stringBoard = session.currentStateString(
                    biggestPatternSize - 1, symmetricSpace);
                
                for (Rule rule : rules) {
                    if (rule.match(stringBoard, symmetricSpace)) {
                        progress = true;
                        callback.actionPerformed(null);
                        continue mainLoop;
                    }
                }
            }
        }
        
        StringBuffer buf = new StringBuffer();
        if (session.mistakes.size() > 0) {
            buf.append("game over because of a mistake\npress <- arrow\nto see the rule that caused it...");
        } else if (session.gameOver) {
            buf.append("mines swept!");
        } else {
            buf.append("...no more rules make progress.");
        }
        buf.append("\n\n");
        buf.append("Rule Match Counts:\n");
        for (Rule r : rules) {
            buf.append("" + r.uses + " : " + r.name + "\n");
        }
        session.annotate(buf.toString());
    }
    
    public void save(String filename) throws Exception {
        session.save(filename);
    }
}
